
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "rif_rgb.h"
#include "sif.h"
#include "sif_rgb.h"

extern void	Error();

static void	MinMax(val,min,max)
double		val,*min,*max;
{
	if (val<*min) *min=val; else if (val>*max) *max=val;
}

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nx,Ny,BPS,Nh,Nv,R,G,B,j,i,gray=1;
	Cell	**oR,**oG,**oB,**nR,**nG,**nB,r,g,b;
	double	X,Y,hx,hy,vx,vy,u,h0,v0,h,v,xh,xv,yh,yv,ratio,x,y;
#ifdef	BLI
	int	Nx1,Ny1,x1,x2,y1,y2;
	double	o,xo,yo;

#define NAME	"bli_RGB"
#else
	int	x0,y0;

#define NAME	"affine_RGB"
#endif

	if (argc!=10 && argc!=14)
	Error("usage : " NAME " orgTIFF hx hy vx vy R G B newTIFF\n"
	      "        " NAME " orgTIFF hx hy vx vy h0 v0 Nh Nv R G B newTIFF");

	ReadImageFile_RGB(argv[1],&Nx,&Ny,&BPS,&oR,&oG,&oB,NULL);

	X=(double)Nx; Y=(double)Ny; 
#ifdef	BLI
	Nx1=Nx-1; Ny1=Ny-1;
#endif
	hx=atof(argv[2]); hy=atof(argv[3]);
	vx=atof(argv[4]); vy=atof(argv[5]); u=hx*vy-hy*vx;
	if (-0.5<X*Y*u && X*Y*u<0.5) Error("bad matrix element.");

	if (argc==10) {
	    h0=v0=h=v=0.0;
	    MinMax(hx*X,&h0,&h); MinMax(vx*X,&v0,&v);
	    MinMax(hy*Y,&h0,&h); MinMax(vy*Y,&v0,&v);
	    MinMax(hx*X+hy*Y,&h0,&h); MinMax(vx*X+vy*Y,&v0,&v);

	    Nh=(int)(h-h0+0.5);
	    Nv=(int)(v-v0+0.5);
	}
	else {
	    h0=atof(argv[6]);
	    v0=atof(argv[7]);
	    if ((Nh=atoi(argv[8]))<=0 ||
		(Nv=atoi(argv[9]))<=0) Error("bad image size.");
	}
	if ((R=atoi(argv[argc-4]))<0 || R>=256 ||
	    (G=atoi(argv[argc-3]))<0 || G>=256 ||
	    (B=atoi(argv[argc-2]))<0 || B>=256)
	    Error("bad R, G and/or B value for space filling.");

	if ((nR=(Cell **)malloc(sizeof(Cell *)*(size_t)Nv))==NULL ||
	    (nG=(Cell **)malloc(sizeof(Cell *)*(size_t)Nv))==NULL ||
	    (nB=(Cell **)malloc(sizeof(Cell *)*(size_t)Nv))==NULL ||
	    (*nR=(Cell *)malloc(sizeof(Cell)*(size_t)Nv*(size_t)Nh))==NULL ||
	    (*nG=(Cell *)malloc(sizeof(Cell)*(size_t)Nv*(size_t)Nh))==NULL ||
	    (*nB=(Cell *)malloc(sizeof(Cell)*(size_t)Nv*(size_t)Nh))==NULL)
	    Error("cell memory allocation error.");

	h0+=0.5; v0+=0.5; xh=vy/u; xv=(-hy)/u;
			  yh=(-vx)/u; yv=hx/u; ratio=255.0/(double)((1<<BPS)-1);

	for (vx=xv*v0, vy=yv*v0, j=0; j<Nv; j++, vx+=xv, vy+=yv) {
	    if (j!=0) {
		nR[j]=nR[j-1]+Nh; nG[j]=nG[j-1]+Nh; nB[j]=nB[j-1]+Nh;
	    }
	    for (hx=xh*h0, hy=yh*h0, i=0; i<Nh; i++, hx+=xh, hy+=yh) {
		if ((x=hx+vx)<0.0 || x>=X || (y=hy+vy)<0.0 || y>=Y) {
		    r=nR[j][i]=(Cell)R; g=nG[j][i]=(Cell)G; b=nB[j][i]=(Cell)B;
		}
		else {

#define OR	(double)oR
#define OG	(double)oG
#define OB	(double)oB

#ifdef	BLI
#define BOUNDARY(t,T,t1,t2,Nt1)	if ((t+=0.5)>=T) { \
				    t1=t2=Nt1; t-=T; \
				} \
				else if (t>=1.0) { \
				    t2=(int)t; t1=t2-1; t-=(double)t2; \
				} \
				else \
				    t1=t2=0

		    BOUNDARY(x,X,x1,x2,Nx1); BOUNDARY(y,Y,y1,y2,Ny1);

	    o=OR[y1][x1]; xo=OR[y1][x2]-o; yo=OR[y2][x1]-o;
	    r=nR[j][i]=(Cell)(ratio*((OR[y2][x2]-xo-yo-o)*x*y+xo*x+yo*y+o)+0.5);
	    o=OG[y1][x1]; xo=OG[y1][x2]-o; yo=OG[y2][x1]-o;
	    g=nG[j][i]=(Cell)(ratio*((OG[y2][x2]-xo-yo-o)*x*y+xo*x+yo*y+o)+0.5);
	    o=OB[y1][x1]; xo=OB[y1][x2]-o; yo=OB[y2][x1]-o;
	    b=nB[j][i]=(Cell)(ratio*((OB[y2][x2]-xo-yo-o)*x*y+xo*x+yo*y+o)+0.5);
#else
		    x0=(int)x; y0=(int)y;
		    r=nR[j][i]=(Cell)(ratio*OR[y0][x0]+0.5);
		    g=nG[j][i]=(Cell)(ratio*OG[y0][x0]+0.5);
		    b=nB[j][i]=(Cell)(ratio*OB[y0][x0]+0.5);
#endif
		}
		if (gray) gray=r==g && r==b;
	    }
	}
	if (gray)
	    StoreImageFile(argv[argc-1],Nh,Nv,8,nR,NULL);
	else
	    StoreImageFile_RGB(argv[argc-1],Nh,Nv,8,nR,nG,nB,NULL);

	return 0;
}
