
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "rif.h"
#include "sif.h"

extern void	Error();

static void	MinMax(val,min,max)
double		val,*min,*max;
{
	if (val<*min) *min=val; else if (val>*max) *max=val;
}

int	main(argc,argv)
int	argc;
char	**argv;
{
	int	Nx,Ny,BPS,Nh,Nv,PV,j,i;
	Cell	**org,**new;
	double	X,Y,hx,hy,vx,vy,u,h0,v0,h,v,xh,xv,yh,yv,x,y;
#ifdef	BLI
	int	Nx1,Ny1,x1,x2,y1,y2;
	double	o,xo,yo;

#define NAME	"bli_gray"
#else
#define NAME	"affine_gray"
#endif

	if (argc!=8 && argc!=12)
	Error("usage : " NAME " orgTIFF hx hy vx vy PV newTIFF\n"
	      "        " NAME " orgTIFF hx hy vx vy h0 v0 Nh Nv PV newTIFF");

	ReadImageFile(argv[1],&Nx,&Ny,&BPS,&org,NULL);

	X=(double)Nx; Y=(double)Ny; 
#ifdef	BLI
	Nx1=Nx-1; Ny1=Ny-1;
#endif
	hx=atof(argv[2]); hy=atof(argv[3]);
	vx=atof(argv[4]); vy=atof(argv[5]); u=hx*vy-hy*vx;
	if (-0.5<X*Y*u && X*Y*u<0.5) Error("bad matrix element.");

	if (argc==8) {
	    h0=v0=h=v=0.0;
	    MinMax(hx*X,&h0,&h); MinMax(vx*X,&v0,&v);
	    MinMax(hy*Y,&h0,&h); MinMax(vy*Y,&v0,&v);
	    MinMax(hx*X+hy*Y,&h0,&h); MinMax(vx*X+vy*Y,&v0,&v);

	    Nh=(int)(h-h0+0.5);
	    Nv=(int)(v-v0+0.5);
	}
	else {
	    h0=atof(argv[6]);
	    v0=atof(argv[7]);
	    if ((Nh=atoi(argv[8]))<=0 ||
		(Nv=atoi(argv[9]))<=0) Error("bad image size.");
	}
	if ((PV=atoi(argv[argc-2]))<0 || PV>=1<<8*sizeof(Cell))
	    Error("bad pixel value for space filling.");

	if ((new=(Cell **)malloc(sizeof(Cell *)*(size_t)Nv))==NULL ||
	    (*new=(Cell *)malloc(sizeof(Cell)*(size_t)Nv*(size_t)Nh))==NULL)
	    Error("cell memory allocation error.");

	h0+=0.5; v0+=0.5; xh=vy/u; xv=(-hy)/u; yh=(-vx)/u; yv=hx/u;

	for (vx=xv*v0, vy=yv*v0, j=0; j<Nv; j++, vx+=xv, vy+=yv) {
	    if (j!=0) new[j]=new[j-1]+Nh;

	    for (hx=xh*h0, hy=yh*h0, i=0; i<Nh; i++, hx+=xh, hy+=yh)
#ifdef	BLI
		if ((x=hx+vx)<0.0 || x>=X || (y=hy+vy)<0.0 || y>=Y)
		    new[j][i]=(Cell)PV;
		else {

#define BOUNDARY(t,T,t1,t2,Nt1)	if ((t+=0.5)>=T) { \
				    t1=t2=Nt1; t-=T; \
				} \
				else if (t>=1.0) { \
				    t2=(int)t; t1=t2-1; t-=(double)t2; \
				} \
				else \
				    t1=t2=0

		    BOUNDARY(x,X,x1,x2,Nx1); BOUNDARY(y,Y,y1,y2,Ny1);

#define O	(double)org

		    o=O[y1][x1]; xo=O[y1][x2]-o; yo=O[y2][x1]-o;
		    new[j][i]=(Cell)((O[y2][x2]-xo-yo-o)*x*y+xo*x+yo*y+o+0.5);
		}
#else
		new[j][i]=((x=hx+vx)<0.0 || x>=X ||
			   (y=hy+vy)<0.0 || y>=Y)?(Cell)PV:org[(int)y][(int)x];
#endif
	}
	while (PV>>BPS) BPS<<=1;
	StoreImageFile(argv[argc-1],Nh,Nv,BPS,new,NULL);

	return 0;
}
