
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "rif.h"
#include "csi.h"
#include "sif.h"

extern void	Error(),ReadSliceImage();

static Cell	***org;

#define ALLOC(type,noe)	(type *)malloc(sizeof(type)*(noe))

static void	Read(Nx,Ny,Nz,paths)
int		Nx,Ny,Nz;
char		**paths;
{
	int	z,y;

	if ((org=ALLOC(Cell **,(size_t)Nz))==NULL ||
	    (*org=ALLOC(Cell *,(size_t)Nz*(size_t)Ny))==NULL ||
	    (**org=ALLOC(Cell,(size_t)Nz*(size_t)Ny*(size_t)Nx))==NULL)
	    Error("no allocatabe memory for org.");

	for (z=0; z<Nz; z++) {
	    if (z==0)
		for (y=1; y<Ny; y++) org[z][y]=org[z][y-1]+Nx;
	    else {
		org[z]=org[z-1]+Ny;
		for (y=0; y<Ny; y++) org[z][y]=org[z][y-1]+Nx;
	    }
	    ReadSliceImage(paths[z],Nx,Ny,org[z]);
	}
}

static double	hx,hy,hz,vx,vy,vz,dx,dy,dz,X,Y,Z,xh,xv,xd,yh,yv,yd,zh,zv,zd;

#define ERROR(msg,line)	(void)fputs(msg,stderr); Error(line)

static void	Matrix(line)
char		*line;
{
	double	u=hx*vy*dz+vx*dy*hz+dx*hy*vz-dx*vy*hz-vx*hy*dz-hx*dy*vz;

	if (-0.5<X*Y*Z*u && X*Y*Z*u<0.5)
	    if (line!=NULL) {
		ERROR("bad matrix element :\n",line);
	    }
	    else
		Error("bad matrix element.");

	xh=(vy*dz-vz*dy)/u; xv=(dy*hz-dz*hy)/u; xd=(hy*vz-hz*vy)/u;
	yh=(vz*dx-vx*dz)/u; yv=(dz*hx-dx*hz)/u; yd=(hz*vx-hx*vz)/u;
	zh=(vx*dy-vy*dx)/u; zv=(dx*hy-dy*hx)/u; zd=(hx*vy-hy*vx)/u;
}

static void	MinMax(val,min,max)
double		val,*min,*max;
{
	if (val<*min) *min=val; else if (val>*max) *max=val;
}

static int	Nh,Nv;
static double	h0,v0,d0;

static int	Area()
{
	double	h,v,d;
	int	Nd;

	h0=v0=d0=h=v=d=0.0;
	MinMax(hx*X,&h0,&h);
	MinMax(vx*X,&v0,&v);
	MinMax(dx*X,&d0,&d);
	MinMax(hy*Y,&h0,&h);
	MinMax(vy*Y,&v0,&v);
	MinMax(dy*Y,&d0,&d);
	MinMax(hz*Z,&h0,&h);
	MinMax(vz*Z,&v0,&v);
	MinMax(dz*Z,&d0,&d);
	MinMax(hx*X+hy*Y,&h0,&h);
	MinMax(vx*X+vy*Y,&v0,&v);
	MinMax(dx*X+dy*Y,&d0,&d);
	MinMax(hy*Y+hz*Z,&h0,&h);
	MinMax(vy*Y+vz*Z,&v0,&v);
	MinMax(dy*Y+dz*Z,&d0,&d);
	MinMax(hz*Z+hx*X,&h0,&h);
	MinMax(vz*Z+vx*X,&v0,&v);
	MinMax(dz*Z+dx*X,&d0,&d);
	MinMax(hx*X+hy*Y+hz*Z,&h0,&h);
	MinMax(vx*X+vy*Y+vz*Z,&v0,&v);
	MinMax(dx*X+dy*Y+dz*Z,&d0,&d);
	Nh=(int)(h-h0+0.5); h0+=0.5;
	Nv=(int)(v-v0+0.5); v0+=0.5;
	Nd=(int)(d-d0+0.5); d0+=0.5; return Nd;
}

static Cell	**new;

static void	Alloc(line)
char		*line;
{
	int	v;

	if ((new=ALLOC(Cell *,(size_t)Nv))==NULL ||
	    (*new=ALLOC(Cell,(size_t)Nv*(size_t)Nh))==NULL)
	    if (line!=NULL) {
		ERROR("no allocatable memory for new :\n",line);
	    }
	    else
		Error("no allocatable memory for new.");

	for (v=1; v<Nv; v++) new[v]=new[v-1]+Nh;
}

static int	BPS,PV;

#ifdef	TLI
static int	Nx1,Ny1,Nz1;
#endif

static void	Store(k,path)
int		k;
char		*path;
{
	double	xvv,yvv,zvv,xhh,yhh,zhh,x,y,z,
		d=d0+(double)k,xdd=xd*d,ydd=yd*d,zdd=zd*d;
	int	v,h,bps;
#ifdef	TLI
	int	x1,x2,y1,y2,z1,z2;
	double	o,xo,yo,zo,xy,yz,zx;
#endif

for (xvv=xv*v0, yvv=yv*v0, zvv=zv*v0, v=0; v<Nv; v++, xvv+=xv, yvv+=yv, zvv+=zv)
for (xhh=xh*h0, yhh=yh*h0, zhh=zh*h0, h=0; h<Nh; h++, xhh+=xh, yhh+=yh, zhh+=zh)
#ifdef	TLI
    if ((x=xhh+xvv+xdd)<0.0 || x>=X ||
	(y=yhh+yvv+ydd)<0.0 || y>=Y ||
	(z=zhh+zvv+zdd)<0.0 || z>=Z)
	new[v][h]=(Cell)PV;
    else {

#define BOUNDARY(t,T,t1,t2,Nt1)	if ((t+=0.5)>=T) { \
				    t1=t2=Nt1; t-=T; \
				} \
				else if (t>=1.0) { \
				    t2=(int)t; t1=t2-1; t-=(double)t2; \
				} \
				else \
				    t1=t2=0

	BOUNDARY(x,X,x1,x2,Nx1);
	BOUNDARY(y,Y,y1,y2,Ny1);
	BOUNDARY(z,Z,z1,z2,Nz1);

#define O	(double)org

	o=O[z1][y1][x1]; xo=O[z1][y1][x2]-o;
			 yo=O[z1][y2][x1]-o; xy=O[z1][y2][x2]-xo-yo-o;
			 zo=O[z2][y1][x1]-o; yz=O[z2][y2][x1]-yo-zo-o;
					     zx=O[z2][y1][x2]-zo-xo-o;
	new[v][h]=(Cell)((O[z2][y2][x2]-xy-yz-zx-xo-yo-zo-o)*x*y*z+
			 xy*x*y+yz*y*z+zx*z*x+xo*x+yo*y+zo*z+o+0.5);
    }
#else
    new[v][h]=((x=xhh+xvv+xdd)<0.0 || x>=X ||
	       (y=yhh+yvv+ydd)<0.0 || y>=Y ||
	       (z=zhh+zvv+zdd)<0.0 || z>=Z)?(Cell)PV:
					    org[(int)z][(int)y][(int)x];
#endif
	for (bps=BPS; PV>>bps; bps<<=1) ;
	StoreImageFile(path,Nh,Nv,bps,new,NULL);
}

#define LEN	1024

int	main(argc,argv)
int	argc;
char	**argv;
{
	char	**paths,line[LEN],str[9][LEN],*ptr[3],path[LEN],form[LEN],
		*bis="bad image size.",
		*bpvfsf="bad pixel value for space filling.";
	int	Nx,Ny,Nz,l,k,Nd;

#ifdef	TLI
#define NAME	"si_tli"
#define PAD	"      "
#else
#define NAME	"si_affine"
#define PAD	"         "
#endif

	if (argc!=3 && argc!=12 && argc!=14 && argc!=15 && argc!=20 && argc!=21)
	Error(
	"usage : " NAME " orgDir nameFile\n"
	"        " NAME " orgDir nameFile hx hy hz vx vy vz dx dy dz\n"
	"        " NAME " orgDir nameFile hx hy hz vx vy vz dx dy dz"
					" PV newDir\n"
	"        " NAME " orgDir nameFile hx hy hz vx vy vz dx dy dz"
					" PV sliceNo newTIFF\n"
	"        " NAME " orgDir nameFile hx hy hz vx vy vz dx dy dz\n"
	"        " PAD  "                 h0 v0 d0 Nh Nv Nd PV newDir\n"
	"        " NAME " orgDir nameFile hx hy hz vx vy vz dx dy dz\n"
	"        " PAD  "                 h0 v0 d0 Nh Nv Nd PV sliceNo newTIFF"
	);

	paths=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,&BPS);

	X=(double)Nx; Y=(double)Ny; Z=(double)Nz;
#ifdef	TLI
	Nx1=Nx-1; Ny1=Ny-1; Nz1=Nz-1;
#endif

switch(argc) {
    case  3:
	Read(Nx,Ny,Nz,paths);
	while (fgets(line,LEN,stdin)!=NULL) {
	    for (l=0; line[l]!='\0'; l++) ;

	    if (line[--l]!='\n') {
		ERROR("missing end of line :\n",line);
	    }
	    line[l]='\0';
	    switch(sscanf(line,"%lf %lf %lf %lf %lf %lf %lf %lf %lf "
			       "%s %s %s %s %s %s %s %s %s",
			       &hx,&hy,&hz,&vx,&vy,&vz,&dx,&dy,&dz,
			       str[0],str[1],str[2],
			       str[3],str[4],str[5],
			       str[6],str[7],str[8])) {
		case  0:
		    continue;
		case 12:
		    Nd=Area();
		    ptr[0]=str[0]; ptr[1]=str[1]; ptr[2]=str[2]; break;
		case 18:
		    h0=atof(str[0])+0.5;
		    v0=atof(str[1])+0.5;
		    d0=atof(str[2])+0.5;
		    if ((Nh=atoi(str[3]))<=0 ||
			(Nv=atoi(str[4]))<=0 ||
			(Nd=atoi(str[5]))<=0) {
			ERROR("bad image size :\n",line);
		    }
		    ptr[0]=str[6]; ptr[1]=str[7]; ptr[2]=str[8]; break;
		default:
		    ERROR("bad parameters :\n",line);
	    }
	    if ((PV=atoi(ptr[0]))<0 || PV>=1<<8*sizeof(Cell)) {
		ERROR("bad pixel value for space filling :\n",line);
	    }
	    Matrix(line);
	    if ((k=atoi(ptr[1]))<0 || k>=Nd) {
		ERROR("bad slice number :\n",line);
	    }
	    Alloc(line); Store(k,ptr[2]); free(*new); free(new);
	}
	break;

#define HX	atof(argv[3])
#define HY	atof(argv[4])
#define HZ	atof(argv[5])
#define VX	atof(argv[6])
#define VY	atof(argv[7])
#define VZ	atof(argv[8])
#define DX	atof(argv[9])
#define DY	atof(argv[10])
#define DZ	atof(argv[11])

    case 12:
	hx=HX; hy=HY; hz=HZ;
	vx=VX; vy=VY; vz=VZ;
	dx=DX; dy=DY; dz=DZ; Nd=Area();
	(void)printf("%lf\t%lf\t%lf\t%d\t%d\t%d\n",
		     h0-0.5,v0-0.5,d0-0.5,Nh,Nv,Nd); break;

#define MATRIX()	hx=HX; hy=HY; hz=HZ; \
			vx=VX; vy=VY; vz=VZ; \
			dx=DX; dy=DY; dz=DZ; Matrix((char *)NULL)

#define AREA(ac)	if (argc==ac) \
			    Nd=Area(); \
			else { \
			    h0=atof(argv[12])+0.5; \
			    v0=atof(argv[13])+0.5; \
			    d0=atof(argv[14])+0.5; \
			    if ((Nh=atoi(argv[15]))<=0 || \
				(Nv=atoi(argv[16]))<=0 || \
				(Nd=atoi(argv[17]))<=0) Error(bis); \
			}

    case 14:
    case 20:
	if ((PV=atoi(argv[argc-2]))<0 || PV>=1<<8*sizeof(Cell)) Error(bpvfsf);

	MATRIX(); AREA(14); Read(Nx,Ny,Nz,paths); Alloc((char *)NULL);

	l=1; for (k=10; k<Nd; k*=10) ++l;
	(void)sprintf(form,"%s/%%0%dd.tif",argv[argc-1],l);

	for (k=0; k<Nd; k++) {
	    (void)sprintf(path,form,k); Store(k,path);
	}
	break;
    case 15:
    case 21:
	if ((PV=atoi(argv[argc-3]))<0 || PV>=1<<8*sizeof(Cell)) Error(bpvfsf);

	MATRIX(); AREA(15);
	if ((k=atoi(argv[argc-2]))<0 || k>=Nd) Error("bad slice number.");

	Read(Nx,Ny,Nz,paths); Alloc((char *)NULL); Store(k,argv[argc-1]); break;
}
	return 0;
}
