
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "rif.h"
#include "csi.h"
#include "sif.h"

extern void	Error(),ReadSliceImage();

static Cell	***org;

#define ALLOC(type,noe)	(type *)malloc(sizeof(type)*(noe))

static void	Read(Nx,Ny,Nz,paths)
int		Nx,Ny,Nz;
char		**paths;
{
	int	z,y;

	if ((org=ALLOC(Cell **,(size_t)Nz))==NULL ||
	    (*org=ALLOC(Cell *,(size_t)Nz*(size_t)Ny))==NULL ||
	    (**org=ALLOC(Cell,(size_t)Nz*(size_t)Ny*(size_t)Nx))==NULL)
	    Error("no allocatabe memory for org.");

	for (z=0; z<Nz; z++) {
	    if (z==0)
		for (y=1; y<Ny; y++) org[z][y]=org[z][y-1]+Nx;
	    else {
		org[z]=org[z-1]+Ny;
		for (y=0; y<Ny; y++) org[z][y]=org[z][y-1]+Nx;
	    }
	    ReadSliceImage(paths[z],Nx,Ny,org[z]);
	}
}

static double	hx,hy,hz,vx,vy,vz,dx,dy,dz;

#ifndef	M_PI
#define M_PI	3.14159265358979323846
#endif

static void	Matrix(lonDeg,latDeg,rotDeg)
double		lonDeg,latDeg,rotDeg;
{
	double	lon=lonDeg*M_PI/180.0,cosLon=cos(lon),sinLon=sin(lon),
		lat=latDeg*M_PI/180.0,cosLat=cos(lat),sinLat=sin(lat),
		rot=rotDeg*M_PI/180.0,cosRot=cos(rot),sinRot=sin(rot);

	hx=  sinLon*cosRot+cosLon*sinLat*sinRot ;
	hy=(-cosLon*cosRot+sinLon*sinLat*sinRot);
	hz=(			 -cosLat*sinRot);
	vx=(-sinLon*sinRot+cosLon*sinLat*cosRot);
	vy=  cosLon*sinRot+sinLon*sinLat*cosRot ;
	vz=(			 -cosLat*cosRot);
	dx=		   cosLon*cosLat	;
	dy=		   sinLon*cosLat	;
	dz=			  sinLat	;
}

static void	MinMax(val,min,max)
double		val,*min,*max;
{
	if (val<*min) *min=val; else if (val>*max) *max=val;
}

static int	Nh,Nv;
static double	X,Y,Z,h0,v0,d0;

static int	Area()
{
	double	h,v,d;
	int	Nd;

	h0=v0=d0=h=v=d=0.0;
	MinMax(hx*X,&h0,&h);
	MinMax(vx*X,&v0,&v);
	MinMax(dx*X,&d0,&d);
	MinMax(hy*Y,&h0,&h);
	MinMax(vy*Y,&v0,&v);
	MinMax(dy*Y,&d0,&d);
	MinMax(hz*Z,&h0,&h);
	MinMax(vz*Z,&v0,&v);
	MinMax(dz*Z,&d0,&d);
	MinMax(hx*X+hy*Y,&h0,&h);
	MinMax(vx*X+vy*Y,&v0,&v);
	MinMax(dx*X+dy*Y,&d0,&d);
	MinMax(hy*Y+hz*Z,&h0,&h);
	MinMax(vy*Y+vz*Z,&v0,&v);
	MinMax(dy*Y+dz*Z,&d0,&d);
	MinMax(hz*Z+hx*X,&h0,&h);
	MinMax(vz*Z+vx*X,&v0,&v);
	MinMax(dz*Z+dx*X,&d0,&d);
	MinMax(hx*X+hy*Y+hz*Z,&h0,&h);
	MinMax(vx*X+vy*Y+vz*Z,&v0,&v);
	MinMax(dx*X+dy*Y+dz*Z,&d0,&d);
	Nh=(int)(h-h0+0.5); h0+=0.5;
	Nv=(int)(v-v0+0.5); v0+=0.5;
	Nd=(int)(d-d0+0.5); d0+=0.5; return Nd;
}

static Cell	**new;

#define ERROR(msg,line)	(void)fputs(msg,stderr); Error(line)

static void	Alloc(line)
char		*line;
{
	int	v;

	if ((new=ALLOC(Cell *,(size_t)Nv))==NULL ||
	    (*new=ALLOC(Cell,(size_t)Nv*(size_t)Nh))==NULL)
	    if (line!=NULL) {
		ERROR("no allocatable memory for new :\n",line);
	    }
	    else
		Error("no allocatable memory for new.");

	for (v=1; v<Nv; v++) new[v]=new[v-1]+Nh;
}

static int	BPS,PV;

#ifdef	TLI
static int	Nx1,Ny1,Nz1;
#endif

#define xh	hx
#define xv	vx
#define xd	dx
#define yh	hy
#define yv	vy
#define yd	dy
#define zh	hz
#define zv	vz
#define zd	dz

static void	Store(k,path)
int		k;
char		*path;
{
	double	xvv,yvv,zvv,xhh,yhh,zhh,x,y,z,
		d=d0+(double)k,xdd=xd*d,ydd=yd*d,zdd=zd*d;
	int	v,h,bps;
#ifdef	TLI
	int	x1,x2,y1,y2,z1,z2;
	double	o,xo,yo,zo,xy,yz,zx;
#endif

for (xvv=xv*v0, yvv=yv*v0, zvv=zv*v0, v=0; v<Nv; v++, xvv+=xv, yvv+=yv, zvv+=zv)
for (xhh=xh*h0, yhh=yh*h0, zhh=zh*h0, h=0; h<Nh; h++, xhh+=xh, yhh+=yh, zhh+=zh)
#ifdef	TLI
    if ((x=xhh+xvv+xdd)<0.0 || x>=X ||
	(y=yhh+yvv+ydd)<0.0 || y>=Y ||
	(z=zhh+zvv+zdd)<0.0 || z>=Z)
	new[v][h]=(Cell)PV;
    else {

#define BOUNDARY(t,T,t1,t2,Nt1)	if ((t+=0.5)>=T) { \
				    t1=t2=Nt1; t-=T; \
				} \
				else if (t>=1.0) { \
				    t2=(int)t; t1=t2-1; t-=(double)t2; \
				} \
				else \
				    t1=t2=0

	BOUNDARY(x,X,x1,x2,Nx1);
	BOUNDARY(y,Y,y1,y2,Ny1);
	BOUNDARY(z,Z,z1,z2,Nz1);

#define O	(double)org

	o=O[z1][y1][x1]; xo=O[z1][y1][x2]-o;
			 yo=O[z1][y2][x1]-o; xy=O[z1][y2][x2]-xo-yo-o;
			 zo=O[z2][y1][x1]-o; yz=O[z2][y2][x1]-yo-zo-o;
					     zx=O[z2][y1][x2]-zo-xo-o;
	new[v][h]=(Cell)((O[z2][y2][x2]-xy-yz-zx-xo-yo-zo-o)*x*y*z+
			 xy*x*y+yz*y*z+zx*z*x+xo*x+yo*y+zo*z+o+0.5);
    }
#else
    new[v][h]=((x=xhh+xvv+xdd)<0.0 || x>=X ||
	       (y=yhh+yvv+ydd)<0.0 || y>=Y ||
	       (z=zhh+zvv+zdd)<0.0 || z>=Z)?(Cell)PV:
					    org[(int)z][(int)y][(int)x];
#endif
	for (bps=BPS; PV>>bps; bps<<=1) ;
	StoreImageFile(path,Nh,Nv,bps,new,NULL);
}

#define LEN	1024

int	main(argc,argv)
int	argc;
char	**argv;
{
	char	**paths,line[LEN],path[LEN],form[LEN],
		*bis="bad image size.",
		*bpvfsf="bad pixel value for space filling.";
	int	Nx,Ny,Nz,l,k,Nd;
	double	lon,lat,rot;

	if (argc!=3 && argc!=6 && argc!=8 && argc!=9)
	    Error("usage : si_ir orgDir nameFile\n"
		  "        si_ir orgDir nameFile lon lat rot\n"
		  "        si_ir orgDir nameFile lon lat rot PV newDir\n"
		  "        si_ir orgDir nameFile lon lat rot PV sliceNo TIFF");

	paths=CheckSliceImages(argv[1],argv[2],&Nx,&Ny,&Nz,&BPS);

	X=(double)Nx; Y=(double)Ny; Z=(double)Nz;
#ifdef	TLI
	Nx1=Nx-1; Ny1=Ny-1; Nz1=Nz-1;
#endif

switch(argc) {
    case 3:
	Read(Nx,Ny,Nz,paths);
	while (fgets(line,LEN,stdin)!=NULL) {
	    for (l=0; line[l]!='\0'; l++) ;

	    if (line[--l]!='\n') {
		ERROR("missing end of line :\n",line);
	    }
	    line[l]='\0';
	    switch(sscanf(line,"%lf %lf %lf %d %d %s",
			       &lon,&lat,&rot,&PV,&k,path)) {
		case 0:
		    continue;
		case 6:
		    break;
		default:
		    ERROR("bad parameters :\n",line);
	    }
	    if (PV<0 || PV>=1<<8*sizeof(Cell)) {
		ERROR("bad pixel value for space filling :\n",line);
	    }
	    Matrix(lon,lat,rot); Nd=Area();
	    if (k<0 || k>=Nd) {
		ERROR("bad slice number :\n",line);
	    }
	    Alloc(line); Store(k,path); free(*new); free(new);
	}
	break;

#define LON	atof(argv[3])
#define LAT	atof(argv[4])
#define ROT	atof(argv[5])

    case 6:
	Matrix(LON,LAT,ROT); Nd=Area();
	(void)printf("%lf\t%lf\t%lf\n%lf\t%lf\t%lf\n%lf\t%lf\t%lf\n"
		     "%lf\t%lf\t%lf\t%d\t%d\t%d\n",
		     hx,hy,hz,vx,vy,vz,dx,dy,dz,
		     h0-0.5,v0-0.5,d0-0.5,Nh,Nv,Nd); break;
    case 8:
	if ((PV=atoi(argv[6]))<0 || PV>=1<<8*sizeof(Cell)) Error(bpvfsf);

	Matrix(LON,LAT,ROT); Nd=Area();
	Read(Nx,Ny,Nz,paths); Alloc((char *)NULL);

	l=1; for (k=10; k<Nd; k*=10) ++l;
	(void)sprintf(form,"%s/%%0%dd.tif",argv[7],l);

	for (k=0; k<Nd; k++) {
	    (void)sprintf(path,form,k); Store(k,path);
	}
	break;
    case 9:
	if ((PV=atoi(argv[6]))<0 || PV>=1<<8*sizeof(Cell)) Error(bpvfsf);

	Matrix(LON,LAT,ROT); Nd=Area();
	if ((k=atoi(argv[7]))<0 || k>=Nd) Error("bad slice number.");

	Read(Nx,Ny,Nz,paths); Alloc((char *)NULL); Store(k,argv[8]); break;
}
	return 0;
}
